<?php

namespace Drupal\paragraphs_revisions_delete\Form;

use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Paragraphs revisions delete form.
 */
class ParagraphsRevisionsDeleteForm extends FormBase {

  /**
   * Entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The database.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected Connection $database;

  /**
   * Constructs a new SettingsForm object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   */
  public function __construct(
    EntityTypeManagerInterface $entityTypeManager,
    ConfigFactoryInterface $configFactory,
    Connection $database
  ) {
    $this->entityTypeManager = $entityTypeManager;
    $this->configFactory = $configFactory;
    $this->database = $database;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('config.factory'),
      $container->get('database')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'paragraphs_revisions_delete_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(
    array $form,
    FormStateInterface $form_state
  ): array {

    // Get the config for this module.
    $config = $this->configFactory->get('paragraphs_revisions_delete.settings');

    // Get all the paragraph ids.
    $pids = $this->getParagraphIds();

    // The revisions paragraph fieldset.
    $form['revisions'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Paragraphs revisions'),
    ];

    // Description of paragraphs revisions.
    $form['revisions']['description'] = [
      '#markup' => 'Settings for the paragraphs revisions delete can be accessed at <a href="/admin/config/paragraphs-revisions-delete/settings">paragraphs revisions delete settings</a>.',
      '#prefix' => '<p>',
      '#suffix' => '</p>',
      ];

    // The number of paragraphs.
    $form['revisions']['number_of_revisions'] = [
      '#markup' => 'There are ' . count($pids) . ' paragraph(s).',
      '#prefix' => '<p>',
      '#suffix' => '</p>',
    ];

    // Description of paragraphs revisions.
    $form['delete_description'] = [
      '#markup' => 'To delete paragraphs with revisions older than ' . $config->get('number_of_months') . ' month(s), click the submit button below.',
      '#prefix' => '<p>',
      '#suffix' => '</p>',
    ];

    // The form actions.
    $form['actions'] = [
      '#type' => 'actions',
    ];

    // The submit button.
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Submit'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(
    array &$form,
    FormStateInterface $form_state
  ) {

    // Get all the paragraph ids.
    $pids = $this->getParagraphIds();

    // Create and set up the batch builder object.
    $batch = new BatchBuilder();
    $batch->setTitle('Running paragraphs revisions delete batch.')
      ->setFinishCallback([self::class, 'batchFinished'])
      ->setInitMessage('Commencing paragraphs revisions delete')
      ->setProgressMessage('Processing...')
      ->setErrorMessage('An error occurred during processing.');

    // Create chunks of 50 groups.
    $chunks = array_chunk($pids, 50);

    // Process each chunk in the array to operations in the batch process.
    foreach ($chunks as $id => $chunk) {
      $args = [
        $id,
        $chunk,
        count($chunks)
      ];
      $batch->addOperation([self::class, 'batchProcess'], $args);
    }

    // Set the actual batch.
    batch_set($batch->toArray());
  }

  /**
   * Process a batch operation.
   *
   * @param int $batchId
   *   The batch ID.
   * @param array $chunk
   *   The chunk to process.
   * @param array $context
   *   Batch context.
   */
  public static function batchProcess(int $batchId, array $pids, int $max, array &$context): void {

    // If there is no progress, set the variables.
    if (!isset($context['sandbox']['progress'])) {
      $context['sandbox']['progress'] = 0;
      $context['sandbox']['max'] = $max;
    }

    // If there is no deleted variables set, set them.
    if (!isset($context['results']['deleted'])) {
      $context['results']['deleted'] = 0;
      $context['results']['progress'] = 0;
      $context['results']['process'] = 'Paragraphs revisions deleted batch completed';
    }

    // Keep track of progress.
    $context['results']['progress'] += count($pids);

    // Message above progress bar.
    $context['message'] = t('Processing group #@batch_id of group size @batch_size for total @count groups.', [
      '@batch_id' => number_format($batchId),
      '@batch_size' => number_format(count($pids)),
      '@count' => number_format($context['sandbox']['max']),
    ]);

    // Get the config date.
    $config_date = date('Y-m-d', strtotime("-24 months"));

    // Step through each of the paragraph ids, and delete
    // any revisions before the config date.
    foreach ($pids as $pid) {

      // Update the progress.
      $context['sandbox']['progress']++;

      // Reset the array of tables.
      $tables = [];

      // Load the paragraph.
      $paragraph = \Drupal::entityTypeManager()->getStorage('paragraph')->load($pid);

      // Get the current revision, so that we do not delete
      // the current revision.
      $current_revision_id = $paragraph->getRevisionId();

      // Get all the field names from the paragraph.
      $field_names = $paragraph->getFieldDefinitions();

      // Step through each of the field names, and set the
      // name of the table in the array.
      foreach (array_keys($field_names) as $field_name) {

        // If this is an actual field, meaning it starts with
        // field_ then add it to the array.
        if (str_starts_with($field_name, 'field_')) {

          // Get the name of the paragraph revision field table.
          $table_name = 'paragraph_revision__' . $field_name;

          // If the table exists, then add it to the array.
          if (\Drupal::database()->schema()->tableExists($table_name)) {
            $tables[] = $table_name;
          }
        }
      }

      // The query to get the revisions of the paragraph.
      $revision_query = \Drupal::database()->select('paragraphs_item_revision', 'r')
        ->fields('r', ['id', 'revision_id'])
        ->condition('r.id', $pid);

      // Get the results of the revision query.
      $results = $revision_query->execute();

      // Step through the results of the revision query and check
      // if we need to delete this specific revision.
      foreach ($results as $result) {

        // Ensure that are not deleting the current revision.
        if ($result->revision_id !== $current_revision_id) {

          // Load the revision of the paragraph.
          $paragraph_revision = \Drupal::entityTypeManager()
            ->getStorage('paragraph')
            ->loadRevision($result->revision_id);

          // Get the date the revision was created.
          $created = date('Y-m-d', $paragraph_revision->getCreatedTime());

          // If the created date is less than the config date,
          // we can actually delete the revision.
          if ($created < $config_date) {
            $context['results']['deleted']++;

            // Delete the revision.
            \Drupal::entityTypeManager()->getStorage('paragraph')->deleteRevision($result->revision_id);

            // Step through all the tables, and delete the field revisions.
            foreach ($tables as $table) {

              // Delete the field from the paragraph field revisions table.
              \Drupal::database()->delete($table)
                ->condition('revision_id', $result->revision_id)
                ->execute();
            }
          }
        }
      }
    }
  }

  /**
   * Handle batch completion.
   *
   * @param bool $success
   *   TRUE if all batch API tasks were completed successfully.
   * @param array $results
   *   A result array from the batch processing operations.
   * @param array $operations
   *   A list of the operations that had not been completed.
   * @param string $elapsed
   *   Batch.inc kindly provides the elapsed processing time in seconds.
   */
  public static function batchFinished(bool $success, array $results, array $operations, string $elapsed): void {

    // Grab the messenger service, this will be needed if the batch was a
    // success or a failure.
    $messenger = \Drupal::messenger();

    // The success variable was true, which indicates that the batch process
    // was successful ( i.e., no errors occurred).
    if ($success) {

      // Show a success message to the user.
      $messenger->addMessage(
        t('@process processed @count, deleted @deleted in @elapsed.',
          [
            '@process' => $results['process'],
            '@count' => $results['progress'],
            '@deleted' => $results['deleted'],
            '@elapsed' => $elapsed,
          ]
        )
      );

      // Log the batch success.
      \Drupal::logger('batch_form_example')->info(
        '@process processed @count, deleted @deleted in @elapsed.',
        [
          '@process' => $results['process'],
          '@count' => $results['progress'],
          '@deleted' => $results['deleted'],
          '@elapsed' => $elapsed,
        ]
      );
    }
    else {

      // An error occurred. $operations contains the operations that remained
      // unprocessed.
      $error_operation = reset($operations);

      // Pick the last operation and report on what happened.
      if ($error_operation) {

        // Set the error message.
        $message = t(
          'An error occurred while processing %error_operation with arguments: @arguments',
          [
            '%error_operation' => print_r($error_operation[0]),
            '@arguments' => print_r($error_operation[1], TRUE),
         ]
        );

        // Actually apply the error message.
        $messenger->addError($message);
      }
    }
  }

  /**
   * Function to get all the paragraph ids.
   *
   * @return array
   *   Array of paragraph ids.
   */
  private function getParagraphIds(): array {

    // Query to get all the paragraph ids.
    $query = $this->entityTypeManager
      ->getStorage('paragraph')
      ->getQuery()
      ->accessCheck(FALSE);

    return $query->execute();
  }

}

